import discord
import os
import json
import sys

TOKEN = os.getenv("DISCORD_BOT_TOKEN")
CHANNEL_ID = os.getenv("DISCORD_CHANNEL_ID")
GITHUB_URL = os.getenv("GITHUB_URL")

messages = []


if __name__ == "__main__":
    if not len(sys.argv) == 2:
        print("Usage: python send-discord-notification.py <result.json>")
        sys.exit(1)
    if not os.path.isfile(sys.argv[1]):
        print(f"Error: {sys.argv[1]} is not a file")
        sys.exit(1)
    if not TOKEN:
        print("Error: DISCORD_BOT_TOKEN is not set")
        sys.exit(1)

    if not CHANNEL_ID:
        print("Error: DISCORD_CHANNEL_ID is not set")
        sys.exit(1)
    client = discord.Client(intents=discord.Intents.default())

    @client.event
    async def on_ready():
        channel = client.get_channel(int(CHANNEL_ID))
        for message in messages:
            try:
                await channel.send(message)
            except Exception as e:
                print(f"Error: {e}")
                sys.exit(1)
        await client.close()

    message = "## Multiplatform test result:\n"
    with open(sys.argv[1], "r") as f:
        result = json.load(f)
        failures = {}
        stages = []
        for platform in sorted(result.keys()):
            for r in result[platform]:
                if r['stage'] not in stages:
                    stages.append(r['stage'])
                if r['status'] != "✅":
                    failures.setdefault(platform, list()).append(r)
        message += f"- Platforms:\n  - {', '.join(sorted(result.keys()))}\n"
        message += f"- Stages:\n  - {', '.join(stages)}\n"
        # Add only the failures.
        if len(failures) > 0:
            message += "### Failures:\n"
            for platform in sorted(failures.keys()):
                if len(message) >= 1500:
                    messages.append(message)
                    message = ""
                message += f"- **{platform}**:\n"
                for r in failures[platform]:
                    message += f"  - {r['stage']}: {r['status']}\n"
    if GITHUB_URL:
        message += "\n"
        message += f"  [Github]({GITHUB_URL})"
    if message:
        messages.append(message)
    client.run(TOKEN)



from __future__ import annotations

import multiprocessing
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
import torch
from torch_geometric.data.graph_store import EdgeAttr, EdgeLayout, GraphStore

from .connection import Connection

if TYPE_CHECKING:
    import sys

    from torch_geometric.typing import EdgeTensorType

    from .database import Database

    if sys.version_info >= (3, 10):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias

StoreKeyType: TypeAlias = tuple[tuple[str], Any, bool]

REL_BATCH_SIZE = 1000000


@dataclass
class Rel:  # noqa: D101
    edge_type: tuple[str, ...]
    layout: str
    is_sorted: bool
    size: tuple[int, ...]
    materialized: bool = False
    edge_index: EdgeTensorType | None = None


class KuzuGraphStore(GraphStore):  # type: ignore[misc]
    """Graph store compatible with `torch_geometric`."""

    def __init__(self, db: Database, num_threads: int | None = None):
        super().__init__()
        self.db = db
        self.connection: Connection | None = None
        self.store: dict[StoreKeyType, Rel] = {}
        if num_threads is None:
            num_threads = multiprocessing.cpu_count()
        self.num_threads = num_threads
        self.__populate_edge_attrs()

    @staticmethod
    def key(attr: EdgeAttr) -> tuple[tuple[str], Any, bool]:  # noqa: D102
        return attr.edge_type, attr.layout.value, attr.is_sorted

    def _put_edge_index(self, edge_index: EdgeTensorType, edge_attr: EdgeAttr) -> None:
        key = self.key(edge_attr)
        if key in self.store:
            self.store[key].edge_index = edge_index
            self.store[key].materialized = True
            self.store[key].size = edge_attr.size
        else:
            self.store[key] = Rel(key[0], key[1], key[2], edge_attr.size, True, edge_index)

    def _get_edge_index(self, edge_attr: EdgeAttr) -> EdgeTensorType | None:
        if edge_attr.layout.value == EdgeLayout.COO.value:  # noqa: SIM102
            # We always return a sorted COO edge index, if the request is
            # for an unsorted COO edge index, we change the is_sorted flag
            # to True and return the sorted COO edge index.
            if edge_attr.is_sorted is False:
                edge_attr.is_sorted = True

        key = self.key(edge_attr)
        if key in self.store:
            rel = self.store[self.key(edge_attr)]
            if not rel.materialized and rel.layout != EdgeLayout.COO.value:
                msg = "Only COO layout is supported"
                raise ValueError(msg)

            if rel.layout == EdgeLayout.COO.value:
                self.__get_edge_coo_from_database(self.key(edge_attr))
            return rel.edge_index
        else:
            return None

    def _remove_edge_index(self, edge_attr: EdgeAttr) -> None:
        key = self.key(edge_attr)
        if key in self.store:
            del self.store[key]

    def get_all_edge_attrs(self) -> list[EdgeAttr]:
        """Return all EdgeAttr from the store values."""
        return [EdgeAttr(rel.edge_type, rel.layout, rel.is_sorted, rel.size) for rel in self.store.values()]

    def __get_edge_coo_from_database(self, key: StoreKeyType) -> None:
        if not self.connection:
            self.connection = Connection(self.db, self.num_threads)

        rel = self.store[key]
        if rel.layout != EdgeLayout.COO.value:
            msg = "Only COO layout is supported"
            raise ValueError(msg)
        if rel.materialized:
            return

        edge_type = rel.edge_type
        num_edges = self.connection._connection.get_num_rels(edge_type[1])
        result = np.empty(2 * num_edges, dtype=np.int64)
        self.connection._connection.get_all_edges_for_torch_geometric(
            result, edge_type[0], edge_type[1], edge_type[2], REL_BATCH_SIZE
        )
        edge_list = torch.from_numpy(result)
        edge_list = edge_list.reshape((2, edge_list.shape[0] // 2))
        rel.edge_index = edge_list
        rel.materialized = True

    def __populate_edge_attrs(self) -> None:
        if not self.connection:
            self.connection = Connection(self.db, self.num_threads)
        rel_tables = self.connection._get_rel_table_names()
        for rel_table in rel_tables:
            edge_type = (rel_table["src"], rel_table["name"], rel_table["dst"])
            size = self.__get_size(edge_type)
            rel = Rel(edge_type, EdgeLayout.COO.value, True, size, False, None)
            self.store[self.key(EdgeAttr(edge_type, EdgeLayout.COO, True))] = rel

    def __get_size(self, edge_type: tuple[str, ...]) -> tuple[int, int]:
        num_nodes = self.connection._connection.get_num_nodes  # type: ignore[union-attr]
        src_count, dst_count = num_nodes(edge_type[0]), num_nodes(edge_type[-1])
        return (src_count, dst_count)



from __future__ import annotations

import kuzu

from type_aliases import ConnDB


def test_get_next(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")

    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2
    result.rows_as_dict()

    row = result.get_next()
    assert row == {"p.name": "Alice", "p.age": 30}
    row = result.get_next()
    assert row == {"p.name": "Bob", "p.age": 40}


def test_get_all(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")

    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2
    result.rows_as_dict()

    rows = result.get_all()
    assert rows[0] == {"p.name": "Alice", "p.age": 30}
    assert rows[1] == {"p.name": "Bob", "p.age": 40}


def test_get_n(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    conn.execute("CREATE (:person {name: 'Cole', age: 20});")

    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 3
    result.rows_as_dict()

    rows = result.get_n(2)

    assert len(rows) == 2
    assert rows[0] == {"p.name": "Alice", "p.age": 30}
    assert rows[1] == {"p.name": "Bob", "p.age": 40}

    rows = result.get_n(1)

    assert len(rows) == 1
    assert rows[0] == {"p.name": "Cole", "p.age": 20}

    rows = result.get_n(1)

    assert len(rows) == 0


def test_dict_iteration(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2

    result.rows_as_dict()
    rows = list(result)
    assert rows[0] == {"p.name": "Alice", "p.age": 30}
    assert rows[1] == {"p.name": "Bob", "p.age": 40}


def test_dict_iteration_inline(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2

    rows = list(result.rows_as_dict())
    assert rows[0] == {"p.name": "Alice", "p.age": 30}
    assert rows[1] == {"p.name": "Bob", "p.age": 40}


def test_on_off(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    conn.execute("CREATE (:person {name: 'Cole', age: 20});")

    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 3

    row = result.get_next()
    assert row == ["Alice", 30]

    result.rows_as_dict()
    row = result.get_next()
    assert row == {"p.name": "Bob", "p.age": 40}

    result.rows_as_dict(False)
    row = result.get_next()
    assert row == ["Cole", 20]



from __future__ import annotations

import multiprocessing
from typing import TYPE_CHECKING, Any

import numpy as np
import torch
from torch import Tensor
from torch_geometric.data.feature_store import FeatureStore, TensorAttr

from .connection import Connection
from .types import Type

if TYPE_CHECKING:
    from torch_geometric.typing import FeatureTensorType

    from .database import Database


class KuzuFeatureStore(FeatureStore):  # type: ignore[misc]
    """Feature store compatible with `torch_geometric`."""

    def __init__(self, db: Database, num_threads: int | None = None):
        super().__init__()
        self.db = db
        self.connection: Connection = None  # type: ignore[assignment]
        self.node_properties_cache: dict[str, dict[str, Any]] = {}
        if num_threads is None:
            num_threads = multiprocessing.cpu_count()
        self.num_threads = num_threads

    def __get_connection(self) -> Connection:
        if self.connection is None:
            self.connection = Connection(self.db, self.num_threads)
        return self.connection

    def _put_tensor(self, tensor: FeatureTensorType, attr: TensorAttr) -> bool:
        raise NotImplementedError

    def _get_tensor(self, attr: TensorAttr) -> FeatureTensorType | None:
        table_name = attr.group_name
        attr_name = attr.attr_name
        attr_info = self.__get_node_property(table_name, attr_name)
        if (
            attr_info["type"]
            not in [
                Type.INT64.value,
                Type.INT32.value,
                Type.INT16.value,
                Type.DOUBLE.value,
                Type.FLOAT.value,
                Type.BOOL.value,
            ]
        ) or (attr_info["dimension"] > 0 and "shape" not in attr_info):
            return self.__get_tensor_by_query(attr)
        return self.__get_tensor_by_scan(attr)

    def __get_tensor_by_scan(self, attr: TensorAttr) -> FeatureTensorType | None:
        table_name = attr.group_name
        attr_name = attr.attr_name
        indices = attr.index

        if indices is None:
            shape = self.get_tensor_size(attr)
            if indices is None:
                indices = np.arange(shape[0], dtype=np.uint64)
        elif isinstance(indices, slice):
            if indices.step is None or indices.step == 1:
                indices = np.arange(indices.start, indices.stop, dtype=np.uint64)
            else:
                indices = np.arange(indices.start, indices.stop, indices.step, dtype=np.uint64)
        elif isinstance(indices, int):
            indices = np.array([indices])

        if table_name not in self.node_properties_cache:
            self.node_properties_cache[table_name] = self.connection._get_node_property_names(table_name)
        attr_info = self.node_properties_cache[table_name][attr_name]

        flat_dim = 1
        if attr_info["dimension"] > 0:
            for i in range(attr_info["dimension"]):
                flat_dim *= attr_info["shape"][i]
        scan_result = self.connection.database._scan_node_table(
            table_name, attr_name, attr_info["type"], flat_dim, indices, self.num_threads
        )

        if attr_info["dimension"] > 0 and "shape" in attr_info:
            result_shape = (len(indices),) + attr_info["shape"]
            scan_result = scan_result.reshape(result_shape)

        result = torch.from_numpy(scan_result)
        return result

    def __get_tensor_by_query(self, attr: TensorAttr) -> FeatureTensorType | None:
        table_name = attr.group_name
        attr_name = attr.attr_name
        indices = attr.index

        self.__get_connection()

        match_clause = f"MATCH (item:{table_name})"
        return_clause = f"RETURN item.{attr_name}"

        if indices is None:
            where_clause = ""
        elif isinstance(indices, int):
            where_clause = f"WHERE offset(id(item)) = {indices}"
        elif isinstance(indices, slice):
            if indices.step is None or indices.step == 1:
                where_clause = f"WHERE offset(id(item)) >= {indices.start} AND offset(id(item)) < {indices.stop}"
            else:
                where_clause = (
                    f"WHERE offset(id(item)) >= {indices.start} AND offset(id(item)) < {indices.stop} "
                    f"AND (offset(id(item)) - {indices.start}) % {indices.step} = 0"
                )
        elif isinstance(indices, (Tensor, list, np.ndarray, tuple)):
            where_clause = "WHERE"
            for i in indices:
                where_clause += f" offset(id(item)) = {int(i)} OR"
            where_clause = where_clause[:-3]
        else:
            msg = f"Invalid attr.index type: {type(indices)!s}"
            raise ValueError(msg)

        query = f"{match_clause} {where_clause} {return_clause}"
        res = self.connection.execute(query)

        result_list = []
        while res.has_next():
            value_array = res.get_next()
            if len(value_array) == 1:
                value_array = value_array[0]
            result_list.append(value_array)
        try:
            return torch.tensor(result_list)
        except Exception:
            return result_list

    def _remove_tensor(self, attr: TensorAttr) -> bool:
        raise NotImplementedError

    def _get_tensor_size(self, attr: TensorAttr) -> tuple[Any, ...]:
        self.__get_connection()
        length_query = f"MATCH (item:{attr.group_name}) RETURN count(item)"
        res = self.connection.execute(length_query)
        length = res.get_next()[0]
        attr_info = self.__get_node_property(attr.group_name, attr.attr_name)
        if attr_info["dimension"] == 0:
            return (length,)
        else:
            return (length,) + attr_info["shape"]

    def __get_node_property(self, table_name: str, attr_name: str) -> dict[str, Any]:
        if table_name in self.node_properties_cache and attr_name in self.node_properties_cache[table_name]:
            return self.node_properties_cache[table_name][attr_name]
        self.__get_connection()
        if table_name not in self.node_properties_cache:
            self.node_properties_cache[table_name] = self.connection._get_node_property_names(table_name)
        if attr_name not in self.node_properties_cache[table_name]:
            msg = f"Attribute {attr_name} not found in group {table_name}"
            raise ValueError(msg)
        attr_info = self.node_properties_cache[table_name][attr_name]
        return attr_info

    def get_all_tensor_attrs(self) -> list[TensorAttr]:
        """Return all TensorAttr from the table nodes."""
        result_list = []
        self.__get_connection()
        for table_name in self.connection._get_node_table_names():
            if table_name not in self.node_properties_cache:
                self.node_properties_cache[table_name] = self.connection._get_node_property_names(table_name)
            for attr_name in self.node_properties_cache[table_name]:
                if self.node_properties_cache[table_name][attr_name]["type"] in [
                    Type.INT64.value,
                    Type.INT32.value,
                    Type.INT16.value,
                    Type.DOUBLE.value,
                    Type.FLOAT.value,
                    Type.BOOL.value,
                ] and (
                    self.node_properties_cache[table_name][attr_name]["dimension"] == 0
                    or "shape" in self.node_properties_cache[table_name][attr_name]
                ):
                    result_list.append(TensorAttr(table_name, attr_name))
        return result_list



from __future__ import annotations

import datetime
from typing import Any

from pandas import Timedelta, Timestamp
from type_aliases import ConnDB


def test_to_networkx_node(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:person) return p"

    res = conn.execute(query)
    nx_graph = res.get_as_networkx()
    nodes = list(nx_graph.nodes(data=True))
    assert len(nodes) == 8

    ground_truth: dict[str, list[Any]] = {
        "ID": [0, 2, 3, 5, 7, 8, 9, 10],
        "fName": [
            "Alice",
            "Bob",
            "Carol",
            "Dan",
            "Elizabeth",
            "Farooq",
            "Greg",
            "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
        ],
        "gender": [1, 2, 1, 2, 1, 2, 2, 2],
        "isStudent": [True, True, False, False, False, True, False, False],
        "eyeSight": [5.0, 5.1, 5.0, 4.8, 4.7, 4.5, 4.9, 4.9],
        "birthdate": [
            datetime.date(1900, 1, 1),
            datetime.date(1900, 1, 1),
            datetime.date(1940, 6, 22),
            datetime.date(1950, 7, 23),
            datetime.date(1980, 10, 26),
            datetime.date(1980, 10, 26),
            datetime.date(1980, 10, 26),
            datetime.date(1990, 11, 27),
        ],
        "registerTime": [
            Timestamp("2011-08-20 11:25:30"),
            Timestamp("2008-11-03 15:25:30.000526"),
            Timestamp("1911-08-20 02:32:21"),
            Timestamp("2031-11-30 12:25:30"),
            Timestamp("1976-12-23 11:21:42"),
            Timestamp("1972-07-31 13:22:30.678559"),
            Timestamp("1976-12-23 04:41:42"),
            Timestamp("2023-02-21 13:25:30"),
        ],
        "lastJobDuration": [
            Timedelta("1082 days 13:02:00"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("0 days 00:18:00.024000"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("1082 days 13:02:00"),
        ],
        "workedHours": [
            [10, 5],
            [12, 8],
            [4, 5],
            [1, 9],
            [2],
            [3, 4, 5, 6, 7],
            [1],
            [10, 11, 12, 3, 4, 5, 6, 7],
        ],
        "usedNames": [
            ["Aida"],
            ["Bobby"],
            ["Carmen", "Fred"],
            ["Wolfeschlegelstein", "Daniel"],
            ["Ein"],
            ["Fesdwe"],
            ["Grad"],
            ["Ad", "De", "Hi", "Kye", "Orlan"],
        ],
        "courseScoresPerTerm": [
            [[10, 8], [6, 7, 8]],
            [[8, 9], [9, 10]],
            [[8, 10]],
            [[7, 4], [8, 8], [9]],
            [[6], [7], [8]],
            [[8]],
            [[10]],
            [[7], [10], [6, 7]],
        ],
        "grades": [
            [96, 54, 86, 92],
            [98, 42, 93, 88],
            [91, 75, 21, 95],
            [76, 88, 99, 89],
            [96, 59, 65, 88],
            [80, 78, 34, 83],
            [43, 83, 67, 43],
            [77, 64, 100, 54],
        ],
        "_label": [
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
        ],
    }
    for i in range(len(nodes)):
        node_id, node = nodes[i]
        assert node_id == f"{node['_label']}_{node['ID']}"
        for key in ground_truth:
            assert node[key] == ground_truth[key][i]


def test_networkx_undirected(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute("MATCH (p1:person)-[r:knows]->(p2:person) WHERE p1.ID <= 3 RETURN p1, r, p2")

    nx_graph = res.get_as_networkx(directed=False)
    assert not nx_graph.is_directed()

    nodes = list(nx_graph.nodes(data=True))
    assert len(nodes) == 4

    edges = list(nx_graph.edges(data=True))

    ground_truth_p: dict[str, list[Any]] = {
        "ID": [0, 2, 3, 5, 7, 8, 9, 10],
        "fName": [
            "Alice",
            "Bob",
            "Carol",
            "Dan",
            "Elizabeth",
            "Farooq",
            "Greg",
            "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
        ],
        "gender": [1, 2, 1, 2, 1, 2, 2, 2],
        "isStudent": [True, True, False, False, False, True, False, False],
        "eyeSight": [5.0, 5.1, 5.0, 4.8, 4.7, 4.5, 4.9, 4.9],
        "birthdate": [
            datetime.date(1900, 1, 1),
            datetime.date(1900, 1, 1),
            datetime.date(1940, 6, 22),
            datetime.date(1950, 7, 23),
            datetime.date(1980, 10, 26),
            datetime.date(1980, 10, 26),
            datetime.date(1980, 10, 26),
            datetime.date(1990, 11, 27),
        ],
        "registerTime": [
            Timestamp("2011-08-20 11:25:30"),
            Timestamp("2008-11-03 15:25:30.000526"),
            Timestamp("1911-08-20 02:32:21"),
            Timestamp("2031-11-30 12:25:30"),
            Timestamp("1976-12-23 11:21:42"),
            Timestamp("1972-07-31 13:22:30.678559"),
            Timestamp("1976-12-23 04:41:42"),
            Timestamp("2023-02-21 13:25:30"),
        ],
        "lastJobDuration": [
            Timedelta("1082 days 13:02:00"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("0 days 00:18:00.024000"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("1082 days 13:02:00"),
        ],
        "workedHours": [
            [10, 5],
            [12, 8],
            [4, 5],
            [1, 9],
            [2],
            [3, 4, 5, 6, 7],
            [1],
            [10, 11, 12, 3, 4, 5, 6, 7],
        ],
        "usedNames": [
            ["Aida"],
            ["Bobby"],
            ["Carmen", "Fred"],
            ["Wolfeschlegelstein", "Daniel"],
            ["Ein"],
            ["Fesdwe"],
            ["Grad"],
            ["Ad", "De", "Hi", "Kye", "Orlan"],
        ],
        "courseScoresPerTerm": [
            [[10, 8], [6, 7, 8]],
            [[8, 9], [9, 10]],
            [[8, 10]],
            [[7, 4], [8, 8], [9]],
            [[6], [7], [8]],
            [[8]],
            [[10]],
            [[7], [10], [6, 7]],
        ],
        "grades": [
            [96, 54, 86, 92],
            [98, 42, 93, 88],
            [91, 75, 21, 95],
            [76, 88, 99, 89],
            [96, 59, 65, 88],
            [80, 78, 34, 83],
            [43, 83, 67, 43],
            [77, 64, 100, 54],
        ],
        "_label": [
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
        ],
    }
    for node_id, node in nodes:
        assert node_id == f"{node['_label']}_{node['ID']}"

    for _, node in nodes:
        found = False
        for i in range(len(ground_truth_p["ID"])):
            if node["ID"] != ground_truth_p["ID"][i]:
                continue
            found = True
            for key in ground_truth_p:
                assert node[key] == ground_truth_p[key][i]
        assert found

    assert len(edges) == 9
    # This should be a complete graph, so we check if an edge exists between
    # every pair of nodes and that there are no self-loops
    for i in range(len(nodes)):
        assert not nx_graph.has_edge(nodes[i][0], nodes[i][0])
        for j in range(i + 1, len(nodes)):
            assert nx_graph.has_edge(nodes[i][0], nodes[j][0])


def test_networkx_directed(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute("MATCH (p:person)-[r:workAt]->(o:organisation) RETURN p, r, o")

    nx_graph = res.get_as_networkx(directed=True)
    assert nx_graph.is_directed()

    nodes = list(nx_graph.nodes(data=True))
    assert len(nodes) == 5

    edges = list(nx_graph.edges(data=True))

    ground_truth_p: dict[str, list[Any]] = {
        "ID": [3, 5, 7],
        "fName": ["Carol", "Dan", "Elizabeth"],
        "gender": [1, 2, 1],
        "isStudent": [False, False, False],
        "eyeSight": [5.0, 4.8, 4.7],
        "birthdate": [
            datetime.date(1940, 6, 22),
            datetime.date(1950, 7, 23),
            datetime.date(1980, 10, 26),
        ],
        "registerTime": [
            Timestamp("1911-08-20 02:32:21"),
            Timestamp("2031-11-30 12:25:30"),
            Timestamp("1976-12-23 11:21:42"),
        ],
        "lastJobDuration": [
            Timedelta("48 hours 24 minutes 11 seconds"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
        ],
        "workedHours": [[4, 5], [1, 9], [2]],
        "usedNames": [["Carmen", "Fred"], ["Wolfeschlegelstein", "Daniel"], ["Ein"]],
        "courseScoresPerTerm": [[[8, 10]], [[7, 4], [8, 8], [9]], [[6], [7], [8]]],
        "grades": [[91, 75, 21, 95], [76, 88, 99, 89], [96, 59, 65, 88]],
        "_label": ["person", "person", "person"],
    }

    for node_id, node in nodes:
        assert node_id == f"{node['_label']}_{node['ID']}"

    for _, node in nodes:
        if "person" not in node:
            continue
        found = False
        for i in range(len(ground_truth_p["ID"])):
            if node["ID"] != ground_truth_p["ID"][i]:
                continue
            found = True
            for key in ground_truth_p:
                assert node[key] == ground_truth_p[key][i]
        assert found

    ground_truth_o: dict[str, list[Any]] = {
        "ID": [4, 6],
        "name": ["CsWork", "DEsWork"],
        "orgCode": [934, 824],
        "mark": [4.1, 4.1],
        "score": [-100, 7],
        "history": ["2 years 4 days 10 hours", "2 years 4 hours 22 us 34 minutes"],
        "licenseValidInterval": [
            Timedelta(days=9414),
            Timedelta(days=3, seconds=36000, microseconds=100000),
        ],
        "rating": [0.78, 0.52],
        "_label": ["organisation", "organisation"],
    }

    for _, node in nodes:
        if "organisation" not in node:
            continue
        found = False
        for i in range(len(ground_truth_o["ID"])):
            if node["ID"] != ground_truth_o["ID"][i]:
                continue
            found = True
            for key in ground_truth_o:
                assert node[key] == ground_truth_o[key][i]
        assert found

    nodes_dict = dict(nx_graph.nodes(data=True))
    edges = list(nx_graph.edges(data=True))
    assert len(edges) == 3

    years_ground_truth = [2010, 2015, 2015]
    for src, dst, edge in edges:
        assert nodes_dict[dst]["_label"] == "organisation"
        assert nodes_dict[dst]["ID"] in [4, 6]
        assert nodes_dict[src]["_label"] == "person"
        assert nodes_dict[src]["ID"] in [3, 5, 7]
        assert edge["year"] in years_ground_truth

        # If the edge is found, remove it from ground truth so we can check
        # that all edges were found and no extra edges were found
        del years_ground_truth[years_ground_truth.index(edge["year"])]
        del nodes_dict[src]

    assert len(years_ground_truth) == 0
    assert len(nodes_dict) == 2  # Only the organisation node should be left


def test_networkx_optional_match(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute(
        """
                       MATCH (u:User)
                       OPTIONAL MATCH (u)-[f:Follows]->(u1:User)
                       RETURN u, f, u1;
                       """
    )

    nx_graph = res.get_as_networkx(directed=True)
    assert nx_graph.is_directed()

    ground_truth_users = [
        {
            "name": "Adam",
            "age": 30,
        },
        {
            "name": "Karissa",
            "age": 40,
        },
        {
            "name": "Zhang",
            "age": 50,
        },
        {
            "name": "Noura",
            "age": 25,
        },
    ]

    ground_truth_follows = [
        {
            "src": "Adam",
            "dst": "Karissa",
            "since": 2020,
        },
        {
            "src": "Adam",
            "dst": "Zhang",
            "since": 2020,
        },
        {
            "src": "Karissa",
            "dst": "Zhang",
            "since": 2021,
        },
        {
            "src": "Zhang",
            "dst": "Noura",
            "since": 2022,
        },
    ]

    nodes = list(nx_graph.nodes(data=True))
    assert len(nodes) == 4
    for node_id, node in nodes:
        assert node_id == f"User_{node['name']}"
        for user in ground_truth_users:
            if user["name"] == node["name"]:
                for key in user:
                    assert node[key] == user[key]
                assert node["User"]

    for src, dst, data in nx_graph.edges(data=True):
        assert src in nx_graph.nodes
        assert dst in nx_graph.nodes
        src_node = nx_graph.nodes[src]
        dst_node = nx_graph.nodes[dst]
        src_name = src_node["name"]
        dst_name = dst_node["name"]
        for follows in ground_truth_follows:
            if follows["src"] == src_name and follows["dst"] == dst_name:
                assert data["since"] == follows["since"]
                break
        else:
            msg = f"Edge {src_name} -> {dst_name} not found in ground truth"
            raise AssertionError(msg)

    edges = list(nx_graph.edges(data=True))
    assert len(edges) == 4



import os

import pytest

from conftest import ShellTest
from test_helper import KUZU_VERSION, deleteIfExists


def test_database_path(temp_db) -> None:
    # no database path
    test = ShellTest()
    result = test.run()
    result.check_stdout("Opening the database under in-memory mode.")

    # valid database path
    test = ShellTest().add_argument(temp_db).statement('RETURN "databases rule" AS a;')
    result = test.run()
    result.check_stdout("databases rule")


@pytest.mark.parametrize(
    "flag",
    ["-h", "--help"],
)
def test_help(temp_db, flag) -> None:
    # database path not needed
    test = ShellTest().add_argument(flag)
    result = test.run()
    result.check_stdout("Kuzu shell")
    # with database path
    test = ShellTest().add_argument(temp_db).add_argument(flag)
    result = test.run()
    result.check_stdout("Kuzu shell")


@pytest.mark.parametrize(
    "flag",
    ["-d", "--defaultbpsize", "--default_bp_size"],
)
def test_default_bp_size(temp_db, flag) -> None:
    # empty flag argument
    test = ShellTest().add_argument(temp_db).add_argument(flag)
    result = test.run()
    result.check_stderr(
        f"Flag '{flag.replace('-', '')}' requires an argument but received none",
    )

    # flag argument is not a number
    test = ShellTest().add_argument(temp_db).add_argument(flag).add_argument("kuzu")
    result = test.run()
    result.check_stderr("Argument '' received invalid value type 'kuzu'")

    # successful flag
    test = ShellTest().add_argument(temp_db).add_argument(flag).add_argument("1000")
    result = test.run()
    result.check_stdout(f"Opening the database at path: {temp_db} in read-write mode.")


@pytest.mark.parametrize(
    "flag",
    ["--maxdbsize", "--max_db_size"],
)
def test_max_db_size(temp_db, flag) -> None:
    # empty flag argument
    test = ShellTest().add_argument(temp_db).add_argument(flag)
    result = test.run()
    result.check_stderr(
        f"Flag '{flag.replace('-', '')}' requires an argument but received none",
    )

    # flag argument is not a number
    test = ShellTest().add_argument(temp_db).add_argument(flag).add_argument("kuzu")
    result = test.run()
    result.check_stderr("Argument 'max_db_size' received invalid value type 'kuzu'")

    # invalid flag
    test = ShellTest().add_argument(temp_db).add_argument(flag).add_argument("1000")
    result = test.run()
    result.check_stderr(
        "Buffer manager exception: The given max db size should be at least 8388608 bytes."
    )

    # successful flag
    test = ShellTest().add_argument(temp_db).add_argument(flag).add_argument("8388608")
    result = test.run()
    result.check_stdout(f"Opening the database at path: {temp_db} in read-write mode.")


@pytest.mark.parametrize(
    "flag",
    ["--nocompression", "--no_compression"],
)
def test_no_compression(temp_db, flag) -> None:
    test = ShellTest().add_argument(temp_db).add_argument(flag)
    result = test.run()
    result.check_stdout(f"Opening the database at path: {temp_db} in read-write mode.")


@pytest.mark.parametrize(
    "flag",
    ["-r", "--readonly", "--read_only"],
)
def test_read_only(temp_db, flag) -> None:
    # cannot open an empty database in read only mode so initialize database
    test = ShellTest().add_argument(temp_db)
    result = test.run()
    result.check_stdout(f"Opening the database at path: {temp_db} in read-write mode.")

    # test read only
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .statement('RETURN "databases rule" AS a;')
        .statement("CREATE NODE TABLE a(i STRING, PRIMARY KEY(i));")
        .statement('RETURN "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout(f"Opening the database at path: {temp_db} in read-only mode.")
    result.check_stdout("databases rule")
    result.check_stdout(
        "Error: Connection exception: Cannot execute write operations in a read-only database!",
    )
    result.check_stdout("kuzu is cool")


def test_history_path(temp_db, history_path) -> None:
    # empty flag argument
    test = ShellTest().add_argument(temp_db).add_argument("-p")
    result = test.run()
    result.check_stderr("Flag 'p' requires an argument but received none")

    # invalid path
    test = ShellTest().add_argument(temp_db).add_argument("-p").add_argument("///////")
    result = test.run()
    result.check_stderr("Error: failed to open the history file: ///////history.txt")

    # valid path, file doesn't exist
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument("-p")
        .add_argument(history_path)
        .statement('RETURN "databases rule" AS a;')
    )
    result = test.run()
    result.check_stdout("databases rule")
    with open(os.path.join(history_path, "history.txt")) as f:
        assert f.readline() == 'RETURN "databases rule" AS a;\n'

    # valid path, file exists
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument("-p")
        .add_argument(history_path)
        .statement('RETURN "kuzu is cool" AS b;')
    )
    result = test.run()
    with open(os.path.join(history_path, "history.txt")) as f:
        assert f.readline() == 'RETURN "databases rule" AS a;\n'
        assert f.readline() == 'RETURN "kuzu is cool" AS b;\n'

    deleteIfExists(os.path.join(history_path, "history.txt"))


@pytest.mark.parametrize(
    "flag",
    [
        "-v",
        "--version",
    ],
)
def test_version(temp_db, flag) -> None:
    # database path not needed
    test = ShellTest().add_argument(flag)
    result = test.run()
    result.check_stdout(KUZU_VERSION)
    # with database path
    test = ShellTest().add_argument(temp_db).add_argument(flag)
    result = test.run()
    result.check_stdout(KUZU_VERSION)


@pytest.mark.parametrize(
    "flag",
    [
        "-m",
        "--mode",
    ],
)
def test_mode(temp_db, flag) -> None:
    # test default mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("\u2502 a              \u2502 b            \u2502")
    result.check_stdout("\u2502 Databases Rule \u2502 kuzu is cool \u2502")

    # test column mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("column")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a                b")
    result.check_stdout("Databases Rule   kuzu is cool")

    # test csv mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("csv")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a,b")
    result.check_stdout("Databases Rule,kuzu is cool")

    # test csv escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("csv")
        .statement(
            'RETURN "This is a \\"test\\", with commas, \\"quotes\\", and\nnewlines.";'
        )
    )
    result = test.run()
    result.check_stdout('"This is a ""test"", with commas, ""quotes"", and\nnewlines."')

    # test box mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("box")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("\u2502 a              \u2502 b            \u2502")
    result.check_stdout("\u2502 Databases Rule \u2502 kuzu is cool \u2502")

    # test html mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("html")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("<table>")
    result.check_stdout("<tr>")
    result.check_stdout("<th>a</th><th>b</th>")
    result.check_stdout("</tr>")
    result.check_stdout("<tr>")
    result.check_stdout("<td>Databases Rule</td><td>kuzu is cool</td>")
    result.check_stdout("</tr>")
    result.check_stdout("</table>")

    # test html escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("html")
        .statement(
            'RETURN "This is a <test> & \\"example\\" with \'special\' characters." AS a;'
        )
    )
    result = test.run()
    result.check_stdout("<table>")
    result.check_stdout("<tr>")
    result.check_stdout("<th>a</th>")
    result.check_stdout("</tr>")
    result.check_stdout("<tr>")
    result.check_stdout(
        "<td>This is a &lt;test&gt; &amp; &quot;example&quot; with &apos;special&apos; characters.</td>"
    )
    result.check_stdout("</tr>")
    result.check_stdout("</table>")

    # test json mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("json")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout('[\n{"a":"Databases Rule","b":"kuzu is cool"}\n]')

    # test json escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("json")
        .statement(
            'RETURN "This is a \\"test\\" with backslashes \\\\, newlines\n, and tabs \t." AS a;'
        )
    )
    result = test.run()
    result.check_stdout(
        '[\n{"a":"This is a \\"test\\" with backslashes \\\\, newlines\\n, and tabs \\t."}\n]'
    )

    # test jsonlines mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("jsonlines")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout('{"a":"Databases Rule","b":"kuzu is cool"}')

    # test jsonlines escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("jsonlines")
        .statement(
            'RETURN "This is a \\"test\\" with backslashes \\\\, newlines\n, and tabs \t." AS a;'
        )
    )
    result = test.run()
    result.check_stdout(
        '{"a":"This is a \\"test\\" with backslashes \\\\, newlines\\n, and tabs \\t."}'
    )

    # test latex mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("latex")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("\\begin{tabular}{ll}")
    result.check_stdout("\\hline")
    result.check_stdout("a&b\\\\")
    result.check_stdout("\\hline")
    result.check_stdout("Databases Rule&kuzu is cool\\\\")
    result.check_stdout("\\hline")
    result.check_stdout("\\end{tabular}")

    # test latex escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("latex")
        .statement(
            'RETURN "This is a test with special characters: %, $, &, #, _, {, }, ~, ^, \\\\, <, and >." AS a;'
        )
    )
    result = test.run()
    result.check_stdout("\\begin{tabular}{l}")
    result.check_stdout("\\hline")
    result.check_stdout("a\\\\")
    result.check_stdout("\\hline")
    result.check_stdout(
        "This is a test with special characters: \\%, \\$, \\&, \\#, \\_, \\{, \\}, \\textasciitilde{}, \\textasciicircum{}, \\textbackslash{}, \\textless{}, and \\textgreater{}.\\\\"
    )
    result.check_stdout("\\hline")
    result.check_stdout("\\end{tabular}")

    # test line mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("line")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a = Databases Rule")
    result.check_stdout("b = kuzu is cool")

    # test list mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("list")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a|b")
    result.check_stdout("Databases Rule|kuzu is cool")

    # test list escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("list")
        .statement(
            'RETURN "This is a \\"test\\", with vertical bars |, \\"quotes\\", and\nnewlines.";'
        )
    )
    result = test.run()
    result.check_stdout(
        '"This is a ""test"", with vertical bars |, ""quotes"", and\nnewlines."'
    )

    # test markdown mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("markdown")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("| a              | b            |")
    result.check_stdout("| Databases Rule | kuzu is cool |")

    # test table mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("table")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("| a              | b            |")
    result.check_stdout("| Databases Rule | kuzu is cool |")

    # test tsv mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("tsv")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a\tb")
    result.check_stdout("Databases Rule\tkuzu is cool")

    # test tsv escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("tsv")
        .statement(
            'RETURN "This is a \\"test\\", with tabs \t, \\"quotes\\", and\nnewlines.";'
        )
    )
    result = test.run()
    result.check_stdout(
        '"This is a ""test"", with tabs \t, ""quotes"", and\nnewlines."'
    )

    # test trash mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("trash")
        .statement("RETURN RANGE(0, 10) AS a;")
    )
    result = test.run()
    result.check_not_stdout("[0,1,2,3,4,5,6,7,8,9,10]")

    # test mode info
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stderr(
        f"Flag '{flag.replace('-', '')}' requires an argument but received none"
    )

    # test invalid mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument("invalid")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stderr("Error: cannot parse output mode: invalid")


@pytest.mark.parametrize(
    "flag",
    [
        "-s",
        "--nostats",
        "--no_stats",
    ],
)
def test_no_stats(temp_db, flag) -> None:
    # test stats off
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .statement('RETURN "Databases Rule" AS a;')
    )
    result = test.run()
    result.check_not_stdout("(1 tuple)")
    result.check_not_stdout("(1 column)")
    result.check_not_stdout("Time: ")

    # test stats on
    test = ShellTest().add_argument(temp_db).statement('RETURN "Databases Rule" AS a;')
    result = test.run()
    result.check_stdout("(1 tuple)")
    result.check_stdout("(1 column)")
    result.check_stdout("Time: ")


@pytest.mark.parametrize(
    "flag",
    ["-b", "--no_progress_bar", "--noprogressbar"],
)
def test_no_progress_bar(temp_db, flag) -> None:
    # progress bar on by default
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement("CALL current_setting('progress_bar') RETURN *;")
    )
    result = test.run()
    result.check_stdout("True")

    # progress bar off
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .statement("CALL current_setting('progress_bar') RETURN *;")
    )
    result = test.run()
    print(result.stdout)
    print(result.stderr)
    result.check_stdout("False")


@pytest.mark.parametrize(
    "flag",
    [
        "-i",
        "--init",
    ],
)
def test_init(temp_db, init_path, flag) -> None:
    # without init
    test = ShellTest().add_argument(temp_db).statement("CALL show_tables() RETURN *;")
    result = test.run()
    result.check_not_stdout("person")
    result.check_not_stdout("organisation")
    result.check_not_stdout("movies")
    result.check_not_stdout("knows")
    result.check_not_stdout("studyAt")
    result.check_not_stdout("workAt")
    result.check_not_stdout("meets")
    result.check_not_stdout("marries")

    # with init
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument(flag)
        .add_argument(init_path)
        .statement("CALL show_tables() RETURN *;")
    )
    result = test.run()
    result.check_stdout("person")
    result.check_stdout("organisation")
    result.check_stdout("movies")
    result.check_stdout("knows")
    result.check_stdout("studyAt")
    result.check_stdout("workAt")
    result.check_stdout("meets")
    result.check_stdout("marries")


def test_bad_flag(temp_db) -> None:
    # without database path
    test = ShellTest().add_argument("-a")
    result = test.run()
    result.check_stderr("Flag could not be matched: 'a'")

    test = ShellTest().add_argument("--badflag")
    result = test.run()
    result.check_stderr("Flag could not be matched: badflag")

    # with database path
    test = ShellTest().add_argument(temp_db).add_argument("-a")
    result = test.run()
    result.check_stderr("Flag could not be matched: 'a'")

    test = ShellTest().add_argument(temp_db).add_argument("--badflag")
    result = test.run()
    result.check_stderr("Flag could not be matched: badflag")



from __future__ import annotations

import os
import shutil
import subprocess

import pexpect
import pytest

from test_helper import KUZU_EXEC_PATH, KUZU_ROOT, deleteIfExists


def pytest_addoption(parser) -> None:
    parser.addoption(
        "--start-offset",
        action="store",
        type=int,
        help="Skip the first 'n' tests",
    )


def pytest_collection_modifyitems(config, items) -> None:
    start_offset = config.getoption("--start-offset")
    if not start_offset:
        # --skiplist not given in cli, therefore move on
        return

    skipped = pytest.mark.skip(reason="included in --skiplist")
    skipped_items = items[:start_offset]
    for item in skipped_items:
        item.add_marker(skipped)


class TestResult:
    def __init__(self, stdout, stderr, status_code) -> None:
        self.stdout: str | bytes = stdout
        self.stderr: str | bytes = stderr
        self.status_code: int = status_code

    def check_stdout(self, expected: str | list[str] | bytes) -> None:
        if isinstance(expected, list):
            expected = "\n".join(expected)
        assert self.status_code == 0
        assert expected in self.stdout

    def check_not_stdout(self, expected: str | list[str] | bytes) -> None:
        if isinstance(expected, list):
            expected = "\n".join(expected)
        assert self.status_code == 0
        assert expected not in self.stdout

    def check_stderr(self, expected: str) -> None:
        assert expected in self.stderr


class ShellTest:
    def __init__(self) -> None:
        self.shell = KUZU_EXEC_PATH
        self.arguments = [self.shell]
        self.statements: list[str] = []
        self.input = None
        self.output = None
        self.environment = {}
        self.shell_process = None

    def add_argument(self, *args):
        self.arguments.extend(args)
        return self

    def statement(self, stmt):
        self.statements.append(stmt)
        return self

    def query(self, *stmts):
        self.statements.extend(stmts)
        return self

    def input_file(self, file_path):
        self.input = file_path
        return self

    def output_file(self, file_path):
        self.output = file_path
        return self

    # Test Running methods

    def get_command(self, cmd: str) -> list[str]:
        command = self.arguments
        if self.input:
            command += [cmd]
        return command

    def get_input_data(self, cmd: str):
        if self.input:
            with open(self.input, "rb") as f:
                input_data = f.read()
        else:
            input_data = bytearray(cmd, "utf8")
        return input_data

    def get_output_pipe(self):
        output_pipe = subprocess.PIPE
        if self.output:
            output_pipe = open(self.output, "w+")
        return output_pipe

    def get_statements(self):
        statements = []
        for statement in self.statements:
            statements.append(statement)
        return "\n".join(statements)

    def get_output_data(self, res):
        if self.output:
            stdout = open(self.output).read()
        else:
            stdout = res.stdout.decode("utf8").strip()
        stderr = res.stderr.decode("utf8").strip()
        return stdout, stderr

    def run(self):
        statements = self.get_statements()
        command = self.get_command(statements)
        input_data = self.get_input_data(statements)
        output_pipe = self.get_output_pipe()

        my_env = os.environ.copy()
        for key, val in self.environment.items():
            my_env[key] = val

        res = subprocess.run(
            command,
            input=input_data,
            stdout=output_pipe,
            stderr=subprocess.PIPE,
            env=my_env,
            check=False,
        )

        stdout, stderr = self.get_output_data(res)
        return TestResult(stdout, stderr, res.returncode)

    def start(self) -> None:
        command = " ".join(self.arguments)

        my_env = os.environ.copy()
        for key, val in self.environment.items():
            my_env[key] = val

        self.shell_process = pexpect.spawn(command, encoding="utf_8", env=my_env)

    def send_finished_statement(self, stmt: str) -> None:
        if self.shell_process:
            assert self.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0
            self.shell_process.send(stmt)
            assert self.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0

    def send_statement(self, stmt: str) -> None:
        if self.shell_process:
            assert self.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0
            self.shell_process.send(stmt)

    def send_control_statement(self, stmt: str) -> None:
        if self.shell_process:
            assert self.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0
            self.shell_process.sendcontrol(stmt)


@pytest.fixture()
def temp_db(tmp_path):
    shutil.rmtree(tmp_path, ignore_errors=True)
    return str(tmp_path)


@pytest.fixture()
def get_tmp_path(tmp_path):
    return str(tmp_path)


@pytest.fixture()
def history_path():
    path = os.path.join(KUZU_ROOT, "tools", "shell", "test", "files")
    deleteIfExists(os.path.join(path, "history.txt"))
    return path


@pytest.fixture()
def csv_path():
    return os.path.join(KUZU_ROOT, "tools", "shell", "test", "files", "vPerson.csv")


@pytest.fixture()
def init_path():
    return os.path.join(KUZU_ROOT, "tools", "shell", "test", "files", "start.cypher")


@pytest.fixture
def test(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def multiline():
    test = ShellTest()
    test.start()
    return test


@pytest.fixture
def singleline():
    test = ShellTest()
    test.start()
    test.send_finished_statement(":singleline\r")
    assert (
        test.shell_process.expect_exact(["Single line mode enabled", pexpect.EOF]) == 0
    )
    return test



#!/usr/bin/env python3

import os
import shutil
import subprocess
import sys
import tarfile
from tempfile import TemporaryDirectory

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def _get_kuzu_version():
    cmake_file = os.path.abspath(os.path.join(base_dir, "..", "CMakeLists.txt"))
    with open(cmake_file) as f:
        for line in f:
            if line.startswith("project(Kuzu VERSION"):
                raw_version = line.split(" ")[2].strip()
                version_nums = raw_version.split(".")
                if len(version_nums) <= 3:
                    return raw_version
                else:
                    dev_suffix = version_nums[3]
                    version = ".".join(version_nums[:3])
                    version += ".dev%s" % dev_suffix
                    return version


if __name__ == "__main__":
    if len(sys.argv) == 2:
        file_name = sys.argv[1]
    else:
        file_name = "kuzu-%s.tar.gz" % _get_kuzu_version()
    print("Creating %s..." % file_name)

    with TemporaryDirectory() as tempdir:
        subprocess.check_call(
            [
                "git",
                "archive",
                "--format",
                "tar",
                "-o",
                os.path.join(tempdir, "kuzu-source.tar"),
                "HEAD",
            ],
            cwd="../..",
        )

        with tarfile.open(os.path.join(tempdir, "kuzu-source.tar")) as tar:
            tar.extractall(path=os.path.join(tempdir, "kuzu-source"), filter=None)

        os.remove(os.path.join(tempdir, "kuzu-source.tar"))

        # Remove components that are not needed for the pip package
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/dataset"))
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/examples"))
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/benchmark"))
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/logo"))
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/extension"))
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/test"))
        shutil.rmtree(os.path.join(tempdir, "kuzu-source/.github"))

        os.makedirs(os.path.join(tempdir, "kuzu"))
        for path in ["setup.py", "setup.cfg", "MANIFEST.in"]:
            shutil.copy2(path, os.path.join(tempdir, path))
        shutil.copy2("../../LICENSE", os.path.join(tempdir, "LICENSE"))
        shutil.copy2("../../README.md", os.path.join(tempdir, "README.md"))

        shutil.copy2(
            "../../tools/python_api/pyproject.toml",
            os.path.join(tempdir, "pyproject.toml"),
        )
        # Update the version in pyproject.toml
        with open(os.path.join(tempdir, "pyproject.toml"), "r") as f:
            lines = f.readlines()
        with open(os.path.join(tempdir, "pyproject.toml"), "w") as f:
            for line in lines:
                if line.startswith("version ="):
                    f.write('version = "%s"\n' % _get_kuzu_version())
                else:
                    f.write(line)
        shutil.copy2("README.md", os.path.join(tempdir, "README_PYTHON_BUILD.md"))
        subprocess.check_call([sys.executable, "setup.py", "egg_info"], cwd=tempdir)
        shutil.copy2(
            os.path.join(tempdir, "kuzu.egg-info", "PKG-INFO"),
            os.path.join(tempdir, "PKG-INFO"),
        )
        with tarfile.open(file_name, "w:gz") as sdist:
            sdist.add(tempdir, "sdist")



import os
import sys
from enum import Enum
from pathlib import Path

KUZU_ROOT = Path(__file__).parent.parent.parent.parent
if sys.platform == "win32":
    # \ in paths is not supported by kuzu's parser
    KUZU_ROOT = str(KUZU_ROOT).replace("\\", "/")

KUZU_EXEC_PATH = os.path.join(
    KUZU_ROOT,
    "build",
    "release",
    "tools",
    "shell",
    "kuzu",
)


def _get_kuzu_version():
    cmake_file = os.path.join(KUZU_ROOT, "CMakeLists.txt")
    with open(cmake_file) as f:
        for line in f:
            if line.startswith("project(Kuzu VERSION"):
                return line.split(" ")[2].strip()
        return None


KUZU_VERSION = _get_kuzu_version()


class KEY_ACTION(Enum):
    KEY_NULL = "\0"  # NULL
    CTRL_A = "a"  # Ctrl-a
    CTRL_B = "b"  # Ctrl-b
    CTRL_C = "c"  # Ctrl-c
    CTRL_D = "d"  # Ctrl-d
    CTRL_E = "e"  # Ctrl-e
    CTRL_F = "f"  # Ctrl-f
    CTRL_G = "g"  # Ctrl-g
    CTRL_H = chr(8)  # Ctrl-h
    TAB = "\t"  # Tab
    CTRL_K = "k"  # Ctrl-k
    CTRL_L = "l"  # Ctrl-l
    ENTER = "\r"  # Enter
    CTRL_N = "n"  # Ctrl-n
    CTRL_P = "p"  # Ctrl-p
    CTRL_R = "r"  # Ctrl-r
    CTRL_S = "s"  # Ctrl-s
    CTRL_T = "t"  # Ctrl-t
    CTRL_U = "u"  # Ctrl-u
    CTRL_W = chr(23)  # Ctrl-w
    ESC = "\27"  # Escape
    BACKSPACE = chr(127)  # Backspace


def deleteIfExists(file) -> None:
    if os.path.exists(file):
        os.remove(file)



from __future__ import annotations

import datetime
import math
from decimal import Decimal
from typing import Any
from uuid import UUID

import kuzu
import pytz
from pandas import Timedelta, Timestamp
from type_aliases import ConnDB


def test_to_df(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly

    def _test_person_to_df(conn: kuzu.Connection) -> None:
        query = "MATCH (p:person) return p.* ORDER BY p.ID"
        pd = conn.execute(query).get_as_df()
        assert pd["p.ID"].tolist() == [0, 2, 3, 5, 7, 8, 9, 10]
        assert str(pd["p.ID"].dtype) == "int64"
        assert pd["p.fName"].tolist() == [
            "Alice",
            "Bob",
            "Carol",
            "Dan",
            "Elizabeth",
            "Farooq",
            "Greg",
            "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
        ]
        assert str(pd["p.fName"].dtype) == "object"
        assert pd["p.gender"].tolist() == [1, 2, 1, 2, 1, 2, 2, 2]
        assert str(pd["p.gender"].dtype) == "int64"
        assert pd["p.isStudent"].tolist() == [
            True,
            True,
            False,
            False,
            False,
            True,
            False,
            False,
        ]
        assert str(pd["p.isStudent"].dtype) == "bool"
        assert pd["p.eyeSight"].tolist() == [5.0, 5.1, 5.0, 4.8, 4.7, 4.5, 4.9, 4.9]
        assert str(pd["p.eyeSight"].dtype) == "float64"
        assert pd["p.birthdate"].tolist() == [
            Timestamp("1900-01-01"),
            Timestamp("1900-01-01"),
            Timestamp("1940-06-22"),
            Timestamp("1950-07-23 "),
            Timestamp("1980-10-26"),
            Timestamp("1980-10-26"),
            Timestamp("1980-10-26"),
            Timestamp("1990-11-27"),
        ]
        assert str(pd["p.birthdate"].dtype) == "datetime64[us]"
        assert pd["p.registerTime"].tolist() == [
            Timestamp("2011-08-20 11:25:30"),
            Timestamp("2008-11-03 15:25:30.000526"),
            Timestamp("1911-08-20 02:32:21"),
            Timestamp("2031-11-30 12:25:30"),
            Timestamp("1976-12-23 11:21:42"),
            Timestamp("1972-07-31 13:22:30.678559"),
            Timestamp("1976-12-23 04:41:42"),
            Timestamp("2023-02-21 13:25:30"),
        ]
        assert str(pd["p.registerTime"].dtype) == "datetime64[us]"
        assert pd["p.lastJobDuration"].tolist() == [
            Timedelta("1082 days 13:02:00"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("0 days 00:18:00.024000"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("1082 days 13:02:00"),
        ]
        assert str(pd["p.lastJobDuration"].dtype) == "timedelta64[ns]"
        assert pd["p.workedHours"].tolist() == [
            [10, 5],
            [12, 8],
            [4, 5],
            [1, 9],
            [2],
            [3, 4, 5, 6, 7],
            [1],
            [10, 11, 12, 3, 4, 5, 6, 7],
        ]
        assert str(pd["p.workedHours"].dtype) == "object"
        assert pd["p.usedNames"].tolist() == [
            ["Aida"],
            ["Bobby"],
            ["Carmen", "Fred"],
            ["Wolfeschlegelstein", "Daniel"],
            ["Ein"],
            ["Fesdwe"],
            ["Grad"],
            ["Ad", "De", "Hi", "Kye", "Orlan"],
        ]
        assert str(pd["p.usedNames"].dtype) == "object"
        assert pd["p.courseScoresPerTerm"].tolist() == [
            [[10, 8], [6, 7, 8]],
            [[8, 9], [9, 10]],
            [[8, 10]],
            [[7, 4], [8, 8], [9]],
            [[6], [7], [8]],
            [[8]],
            [[10]],
            [[7], [10], [6, 7]],
        ]
        assert str(pd["p.courseScoresPerTerm"].dtype) == "object"
        assert pd["p.grades"].tolist() == [
            [96, 54, 86, 92],
            [98, 42, 93, 88],
            [91, 75, 21, 95],
            [76, 88, 99, 89],
            [96, 59, 65, 88],
            [80, 78, 34, 83],
            [43, 83, 67, 43],
            [77, 64, 100, 54],
        ]
        assert str(pd["p.grades"].dtype) == "object"
        expected_values = [1.731, 0.99, 1.00, 1.30, 1.463, 1.51, 1.6, 1.323]
        actual_values = pd["p.height"].tolist()
        for expected, actual in zip(expected_values, actual_values):
            assert math.isclose(actual, expected, rel_tol=1e-5)
        assert str(pd["p.height"].dtype) == "float32"

    def _test_study_at_to_df(conn: kuzu.Connection) -> None:
        query = "MATCH (p:person)-[r:studyAt]->(o:organisation) return r.* order by r.length;"
        pd = conn.execute(query).get_as_df()
        assert pd["r.year"].tolist() == [2021, 2020, 2020]
        assert str(pd["r.year"].dtype) == "int64"
        assert pd["r.places"].tolist() == [
            ["wwAewsdndweusd", "wek"],
            ["awndsnjwejwen", "isuhuwennjnuhuhuwewe"],
            ["anew", "jsdnwusklklklwewsd"],
        ]
        assert str(pd["r.places"].dtype) == "object"
        assert pd["r.length"].tolist() == [5, 22, 55]
        assert str(pd["r.length"].dtype) == "int16"
        assert pd["r.level"].tolist() == [5, 2, 120]
        assert str(pd["r.level"].dtype) == "int8"
        assert pd["r.code"].tolist() == [9223372036854775808, 23, 6689]
        assert str(pd["r.code"].dtype) == "uint64"
        assert pd["r.temperature"].tolist() == [32800, 20, 1]
        assert str(pd["r.temperature"].dtype) == "uint32"
        assert pd["r.ulength"].tolist() == [33768, 180, 90]
        assert str(pd["r.ulength"].dtype) == "uint16"
        assert pd["r.ulevel"].tolist() == [250, 12, 220]
        assert str(pd["r.ulevel"].dtype) == "uint8"
        assert pd["r.hugedata"].tolist() == [
            1.8446744073709552e27,
            -15.0,
            -1.8446744073709552e21,
        ]
        assert str(pd["r.hugedata"].dtype) == "float64"

    def _test_timestamps_to_df(conn: kuzu.Connection) -> None:
        query = (
            'RETURN cast("2012-01-01 11:12:12.12345", "TIMESTAMP_NS") as A, cast("2012-01-01 11:12:12.12345", '
            '"TIMESTAMP_MS") as B, cast("2012-01-01 11:12:12.12345", "TIMESTAMP_SEC") as C, '
            'cast("2012-01-01 11:12:12.12345", "TIMESTAMP_TZ") as D'
        )
        pd = conn.execute(query).get_as_df()
        assert pd["A"].tolist() == [Timestamp("2012-01-01 11:12:12.123450")]
        assert pd["B"].tolist() == [Timestamp("2012-01-01 11:12:12.123000")]
        assert pd["C"].tolist() == [Timestamp("2012-01-01 11:12:12")]
        assert pd["D"].tolist() == [Timestamp("2012-01-01 11:12:12.123450")]

    def _test_movies_to_df(conn: kuzu.Connection) -> None:
        query = "MATCH (m:movies) return m.* order by m.length;"
        pd = conn.execute(query).get_as_df()
        assert pd["m.length"].tolist() == [126, 298, 2544]
        assert str(pd["m.length"].dtype) == "int32"
        assert pd["m.description"].tolist() == [
            {
                "rating": 5.3,
                "stars": 2,
                "views": 152,
                "release": datetime.datetime(2011, 8, 20, 11, 25, 30),
                "release_ns": datetime.datetime(2011, 8, 20, 11, 25, 30, 123456),
                "release_ms": datetime.datetime(2011, 8, 20, 11, 25, 30, 123000),
                "release_sec": datetime.datetime(2011, 8, 20, 11, 25, 30),
                "release_tz": datetime.datetime(2011, 8, 20, 11, 25, 30, 123456, pytz.UTC),
                "film": datetime.date(2012, 5, 11),
                "u8": 220,
                "u16": 20,
                "u32": 1,
                "u64": 180,
                "hugedata": Decimal("1844674407370955161811111111"),
            },
            {
                "rating": 1223.0,
                "stars": 100,
                "views": 10003,
                "release": datetime.datetime(2011, 2, 11, 16, 44, 22),
                "release_ns": datetime.datetime(2011, 2, 11, 16, 44, 22, 123456),
                "release_ms": datetime.datetime(2011, 2, 11, 16, 44, 22, 123000),
                "release_sec": datetime.datetime(2011, 2, 11, 16, 44, 22),
                "release_tz": datetime.datetime(2011, 2, 11, 16, 44, 22, 123456, pytz.UTC),
                "film": datetime.date(2013, 2, 22),
                "u8": 1,
                "u16": 15,
                "u32": 200,
                "u64": 4,
                "hugedata": Decimal(-15),
            },
            {
                "rating": 7.0,
                "stars": 10,
                "views": 982,
                "release": datetime.datetime(2018, 11, 13, 13, 33, 11),
                "release_ns": datetime.datetime(2018, 11, 13, 13, 33, 11, 123456),
                "release_ms": datetime.datetime(2018, 11, 13, 13, 33, 11, 123000),
                "release_sec": datetime.datetime(2018, 11, 13, 13, 33, 11),
                "release_tz": datetime.datetime(2018, 11, 13, 13, 33, 11, 123456, pytz.UTC),
                "film": datetime.date(2014, 9, 12),
                "u8": 12,
                "u16": 120,
                "u32": 55,
                "u64": 1,
                "hugedata": Decimal(-1844674407370955161511),
            },
        ]
        assert str(pd["m.description"].dtype) == "object"
        assert pd["m.content"].tolist() == [
            b"\xaa\xabinteresting\x0b",
            b"pure ascii characters",
            b"\xab\xcd",
        ]
        assert str(pd["m.content"].dtype) == "object"
        assert pd["m.audience"].tolist() == [
            {"audience1": 52, "audience53": 42},
            {},
            {"audience1": 33},
        ]
        assert str(pd["m.audience"].dtype) == "object"
        assert pd["m.grade"].tolist() == [True, 254.0, 8.989]
        assert str(pd["m.grade"].dtype) == "object"

    def _test_serial_to_df(conn: kuzu.Connection) -> None:
        df = conn.execute("MATCH (a:moviesSerial) RETURN a.ID AS id").get_as_df()
        assert len(df) == 3
        assert df["id"].tolist() == [0, 1, 2]

    _test_person_to_df(conn)
    conn.set_max_threads_for_exec(2)
    _test_study_at_to_df(conn)
    _test_movies_to_df(conn)
    _test_timestamps_to_df(conn)
    _test_serial_to_df(conn)


def test_df_multiple_times(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:person) return p.ID ORDER BY p.ID"
    res = conn.execute(query)
    df = res.get_as_df()
    df_2 = res.get_as_df()
    df_3 = res.get_as_df()
    assert df["p.ID"].tolist() == [0, 2, 3, 5, 7, 8, 9, 10]
    assert df_2["p.ID"].tolist() == [0, 2, 3, 5, 7, 8, 9, 10]
    assert df_3["p.ID"].tolist() == [0, 2, 3, 5, 7, 8, 9, 10]


def test_df_get_node(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:person) return p"

    res = conn.execute(query)
    df = res.get_as_df()
    p_list = df["p"].tolist()
    assert len(p_list) == 8

    ground_truth: dict[str, list[Any]] = {
        "ID": [0, 2, 3, 5, 7, 8, 9, 10],
        "fName": [
            "Alice",
            "Bob",
            "Carol",
            "Dan",
            "Elizabeth",
            "Farooq",
            "Greg",
            "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
        ],
        "gender": [1, 2, 1, 2, 1, 2, 2, 2],
        "isStudent": [True, True, False, False, False, True, False, False],
        "eyeSight": [5.0, 5.1, 5.0, 4.8, 4.7, 4.5, 4.9, 4.9],
        "birthdate": [
            datetime.date(1900, 1, 1),
            datetime.date(1900, 1, 1),
            datetime.date(1940, 6, 22),
            datetime.date(1950, 7, 23),
            datetime.date(1980, 10, 26),
            datetime.date(1980, 10, 26),
            datetime.date(1980, 10, 26),
            datetime.date(1990, 11, 27),
        ],
        "registerTime": [
            Timestamp("2011-08-20 11:25:30"),
            Timestamp("2008-11-03 15:25:30.000526"),
            Timestamp("1911-08-20 02:32:21"),
            Timestamp("2031-11-30 12:25:30"),
            Timestamp("1976-12-23 11:21:42"),
            Timestamp("1972-07-31 13:22:30.678559"),
            Timestamp("1976-12-23 04:41:42"),
            Timestamp("2023-02-21 13:25:30"),
        ],
        "lastJobDuration": [
            Timedelta("1082 days 13:02:00"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
            Timedelta("0 days 00:18:00.024000"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("1082 days 13:02:00"),
        ],
        "workedHours": [
            [10, 5],
            [12, 8],
            [4, 5],
            [1, 9],
            [2],
            [3, 4, 5, 6, 7],
            [1],
            [10, 11, 12, 3, 4, 5, 6, 7],
        ],
        "usedNames": [
            ["Aida"],
            ["Bobby"],
            ["Carmen", "Fred"],
            ["Wolfeschlegelstein", "Daniel"],
            ["Ein"],
            ["Fesdwe"],
            ["Grad"],
            ["Ad", "De", "Hi", "Kye", "Orlan"],
        ],
        "courseScoresPerTerm": [
            [[10, 8], [6, 7, 8]],
            [[8, 9], [9, 10]],
            [[8, 10]],
            [[7, 4], [8, 8], [9]],
            [[6], [7], [8]],
            [[8]],
            [[10]],
            [[7], [10], [6, 7]],
        ],
        "_label": [
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
            "person",
        ],
    }
    for i in range(len(p_list)):
        p = p_list[i]
        for key in ground_truth:
            assert p[key] == ground_truth[key][i]


def test_df_get_node_rel(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute("MATCH (p:person)-[r:workAt]->(o:organisation) RETURN p, r, o ORDER BY p.fName")

    df = res.get_as_df()
    p_list = df["p"].tolist()
    o_list = df["o"].tolist()

    assert len(p_list) == 3
    assert len(o_list) == 3

    ground_truth_p: dict[str, list[Any]] = {
        "ID": [3, 5, 7],
        "fName": ["Carol", "Dan", "Elizabeth"],
        "gender": [1, 2, 1],
        "isStudent": [False, False, False],
        "eyeSight": [5.0, 4.8, 4.7],
        "birthdate": [
            datetime.date(1940, 6, 22),
            datetime.date(1950, 7, 23),
            datetime.date(1980, 10, 26),
        ],
        "registerTime": [
            Timestamp("1911-08-20 02:32:21"),
            Timestamp("2031-11-30 12:25:30"),
            Timestamp("1976-12-23 11:21:42"),
        ],
        "lastJobDuration": [
            Timedelta("48 hours 24 minutes 11 seconds"),
            Timedelta("3750 days 13:00:00.000024"),
            Timedelta("2 days 00:24:11"),
        ],
        "workedHours": [[4, 5], [1, 9], [2]],
        "usedNames": [["Carmen", "Fred"], ["Wolfeschlegelstein", "Daniel"], ["Ein"]],
        "courseScoresPerTerm": [[[8, 10]], [[7, 4], [8, 8], [9]], [[6], [7], [8]]],
        "_label": ["person", "person", "person"],
    }
    for i in range(len(p_list)):
        p = p_list[i]
        for key in ground_truth_p:
            assert p[key] == ground_truth_p[key][i]

    ground_truth_o: dict[str, list[Any]] = {
        "ID": [4, 6, 6],
        "name": ["CsWork", "DEsWork", "DEsWork"],
        "orgCode": [934, 824, 824],
        "mark": [4.1, 4.1, 4.1],
        "score": [-100, 7, 7],
        "history": [
            "2 years 4 days 10 hours",
            "2 years 4 hours 22 us 34 minutes",
            "2 years 4 hours 22 us 34 minutes",
        ],
        "licenseValidInterval": [
            Timedelta(days=9414),
            Timedelta(days=3, seconds=36000, microseconds=100000),
            Timedelta(days=3, seconds=36000, microseconds=100000),
        ],
        "rating": [0.78, 0.52, 0.52],
        "_label": ["organisation", "organisation", "organisation"],
    }
    for i in range(len(o_list)):
        o = df["o"][i]
        for key in ground_truth_o:
            assert o[key] == ground_truth_o[key][i]

    assert df["r"][0]["year"] == 2015
    assert df["r"][1]["year"] == 2010
    assert df["r"][2]["year"] == 2015

    for i in range(len(df["r"])):
        assert df["r"][i]["_src"] == df["p"][i]["_id"]
        assert df["r"][i]["_dst"] == df["o"][i]["_id"]


def test_df_get_recursive_join(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute(
        "MATCH (p:person)-[r:knows*1..2 (e, n | WHERE e.comments = ['rnme','m8sihsdnf2990nfiwf'])]-(m:person) WHERE "
        "p.ID = 0 and m.ID = 0 RETURN r"
    ).get_as_df()
    assert res["r"][0] == {
        "_nodes": [
            {
                "ID": 2,
                "_id": {"offset": 1, "table": 0},
                "_label": "person",
                "age": 30,
                "birthdate": datetime.date(1900, 1, 1),
                "courseScoresPerTerm": [[8, 9], [9, 10]],
                "eyeSight": 5.1,
                "fName": "Bob",
                "gender": 2,
                "grades": [98, 42, 93, 88],
                "height": 0.9900000095367432,
                "u": UUID("a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a12"),
                "isStudent": True,
                "isWorker": False,
                "lastJobDuration": datetime.timedelta(days=3750, seconds=46800, microseconds=24),
                "registerTime": datetime.datetime(2008, 11, 3, 15, 25, 30, 526),
                "usedNames": ["Bobby"],
                "workedHours": [12, 8],
            }
        ],
        "_rels": [
            {
                "_id": {
                    "offset": 0,
                    "table": 3,
                },
                "_dst": {"offset": 1, "table": 0},
                "_label": "knows",
                "_src": {"offset": 0, "table": 0},
                "comments": ["rnme", "m8sihsdnf2990nfiwf"],
                "date": datetime.date(2021, 6, 30),
                "meetTime": datetime.datetime(1986, 10, 21, 21, 8, 31, 521000),
                "notes": 1,
                "summary": {
                    "locations": ["'toronto'", "'waterloo'"],
                    "transfer": {
                        "amount": [100, 200],
                        "day": datetime.date(2021, 1, 2),
                    },
                },
                "someMap": {"a": "b"},
                "validInterval": datetime.timedelta(days=3750, seconds=46800, microseconds=24),
            },
            {
                "_id": {
                    "offset": 0,
                    "table": 3,
                },
                "_dst": {"offset": 1, "table": 0},
                "_label": "knows",
                "_src": {"offset": 0, "table": 0},
                "comments": ["rnme", "m8sihsdnf2990nfiwf"],
                "date": datetime.date(2021, 6, 30),
                "meetTime": datetime.datetime(1986, 10, 21, 21, 8, 31, 521000),
                "notes": 1,
                "summary": {
                    "locations": ["'toronto'", "'waterloo'"],
                    "transfer": {
                        "amount": [100, 200],
                        "day": datetime.date(2021, 1, 2),
                    },
                },
                "someMap": {"a": "b"},
                "validInterval": datetime.timedelta(days=3750, seconds=46800, microseconds=24),
            },
        ],
    }


def test_get_df_unicode(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute("MATCH (m:movies) RETURN m.name").get_as_df()
    assert res["m.name"].tolist() == [
        "Sóló cón tu párejâ",
        "The 😂😃🧘🏻‍♂️🌍🌦️🍞🚗 movie",
        "Roma",
    ]


def test_get_df_decimal(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute(
        "UNWIND [1, 2, 3] AS A UNWIND [5.7, 8.3, 2.9] AS B RETURN CAST(CAST(A AS DECIMAL) * CAST(B AS DECIMAL) AS DECIMAL(18, 1)) AS PROD"
    ).get_as_df()
    assert sorted(res["PROD"].tolist()) == sorted([
        Decimal("5.7"),
        Decimal("8.3"),
        Decimal("2.9"),
        Decimal("11.4"),
        Decimal("16.6"),
        Decimal("5.8"),
        Decimal("17.1"),
        Decimal("24.9"),
        Decimal("8.7"),
    ])
    res = conn.execute(
        "UNWIND [1, 2, 3] AS A UNWIND [5.7, 8.3, 2.9] AS B RETURN CAST(CAST(A AS DECIMAL) * CAST(B AS DECIMAL) AS DECIMAL(4, 1)) AS PROD"
    ).get_as_df()
    assert sorted(res["PROD"].tolist()) == sorted([
        Decimal("5.7"),
        Decimal("8.3"),
        Decimal("2.9"),
        Decimal("11.4"),
        Decimal("16.6"),
        Decimal("5.8"),
        Decimal("17.1"),
        Decimal("24.9"),
        Decimal("8.7"),
    ])



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .connection import Connection


class PreparedStatement:
    """
    A prepared statement is a parameterized query which can avoid planning the
    same query for repeated execution.
    """

    def __init__(self, connection: Connection, query: str, parameters: dict[str, Any] | None = None):
        """
        Parameters
        ----------
        connection : Connection
            Connection to a database.
        query : str
            Query to prepare.
        parameters : dict[str, Any]
            Parameters for the query.
        """
        if parameters is None:
            parameters = {}
        self._prepared_statement = connection._connection.prepare(query, parameters)
        self._connection = connection

    def is_success(self) -> bool:
        """
        Check if the prepared statement is successfully prepared.

        Returns
        -------
        bool
            True if the prepared statement is successfully prepared.
        """
        return self._prepared_statement.is_success()

    def get_error_message(self) -> str:
        """
        Get the error message if the query is not prepared successfully.

        Returns
        -------
        str
            Error message.
        """
        return self._prepared_statement.get_error_message()
